rarity_levels = ["magical","rare","mystical","angelic","mythical","arcane","legendary","godly","epic","relic","artifact","unique"]
transmute_mapping = {
    "I" : {
        "rarity_levels" : rarity_levels[0:-5],
        "ranks" :{"min" : 0, "max" : 4}
    },
    "II" : {
        "rarity_levels" : rarity_levels[0:-5],
        "ranks" :{"min" : 5, "max" : 9}
    },
    "III" : {
        "rarity_levels" : rarity_levels[0:-5],
        "ranks" :{"min" : 10, "max" : 14}
    } ,
    "IV" : {
        "rarity_levels" : rarity_levels[0:-5],
        "ranks" :{"min" : 15, "max" : 19}
    },
    "V" : {
        "rarity_levels" : rarity_levels[0:-5],
        "ranks" :{"min" : 20, "max" : 24}
    },
    "VI" : {
        "rarity_levels" : rarity_levels[0:-5],
        "ranks" :{"min" : 25, "max" : 30}
    },
    "VII" : {
        "rarity_levels" : rarity_levels[0:-5],
        "ranks" :{"min" : 30, "max" : 34}
    },
    "VIII" : {
        "rarity_levels" : rarity_levels[0:-5],
        "ranks" :{"min" : 35, "max" : 39}
    },
    "IX" : {
        "rarity_levels" : rarity_levels[0:-5],
        "ranks" :{"min" : 40, "max" : 44}
    },
    "X" : {
        "rarity_levels" : rarity_levels[0:-5],
        "ranks" :{"min" : 45, "max" : 49}
    },
    "XI" : {
        "rarity_levels" : rarity_levels[0:-5],
        "ranks" :{"min" : 50, "max" : 54}
    },
    "XII" : {
        "rarity_levels" : rarity_levels[0:-5],
        "ranks" :{"min" : 55, "max" : 59}
    },
    "XIII" : {
        "rarity_levels" : rarity_levels,
        "ranks" :{"min" : 60, "max" : 70}
    }
}
item_pickup_mapping = {
    "I" : {
        "rarity_levels" : rarity_levels[0:-5],
        "ranks" :{"min" : 0, "max" : 4}
    },
    "II" : {
        "rarity_levels" : rarity_levels[0:-5],
        "ranks" :{"min" : 5, "max" : 9}
    },
    "III" : {
        "rarity_levels" : rarity_levels[0:-5],
        "ranks" :{"min" : 10, "max" : 14}
    } ,
    "IV" : {
        "rarity_levels" : rarity_levels[0:-5],
        "ranks" :{"min" : 15, "max" : 19}
    },
    "V" : {
        "rarity_levels" : rarity_levels[0:-5],
        "ranks" :{"min" : 20, "max" : 24}
    },
    "VI" : {
        "rarity_levels" : rarity_levels[0:-5],
        "ranks" :{"min" : 25, "max" : 30}
    },
    "VII" : {
        "rarity_levels" : rarity_levels[0:-5],
        "ranks" :{"min" : 30, "max" : 34}
    },
    "VIII" : {
        "rarity_levels" : rarity_levels[0:-5],
        "ranks" :{"min" : 35, "max" : 39}
    },
    "IX" : {
        "rarity_levels" : rarity_levels[0:-5],
        "ranks" :{"min" : 40, "max" : 44}
    },
    "X" : {
        "rarity_levels" : rarity_levels[0:-5],
        "ranks" :{"min" : 45, "max" : 49}
    },
    "XI" : {
        "rarity_levels" : rarity_levels[0:-5],
        "ranks" :{"min" : 50, "max" : 54}
    },
    "XII" : {
        "rarity_levels" : rarity_levels[0:-5],
        "ranks" :{"min" : 55, "max" : 59}
    },
    "XIII" : {
        "rarity_levels" : rarity_levels,
        "ranks" :{"min" : 60, "max" : 70}
    }
}
deposit_criteria = {
    "III" : [
        "+50% Enhanced Effect",
        "+15 Strength",
        "+10 Dexterity",
        "+10 Intelligence",
        "+280 Max Mana",
        "+3 Magic Luck",
        "+3 Extra Equipment Slots",
    ],
    "IV" : [
        "+65% Enhanced Effect",
        "+15 Strength",
        "+15 Dexterity",
        "+15 Intelligence",
        "+290 Max Mana",
        "+3 Magic Luck",
        "+4 Extra Equipment Slots",
    ],
    "V" : [
        "+65% Enhanced Effect",
        "+20 Strength",
        "+20 Dexterity",
        "+20 Intelligence",
        "+350 Max Mana",
        "+3 Magic Luck",
        "+5 Extra Equipment Slots",
    ],
    "VI" : [
        "+75% Enhanced Effect",
        "+20 Strength",
        "+20 Dexterity",
        "+20 Intelligence",
        "+350 Max Mana",
        "+3 Magic Luck",
        "+6 Extra Equipment Slots",
    ],
    "VII" : [
        "+80% Enhanced Effect",
        "+20 Strength",
        "+20 Dexterity",
        "+25 Intelligence",
        "+4 Magic Luck",
        "+6 Extra Equipment Slots",
    ],
    "VIII" : [
        "+90% Enhanced Effect",
        "+25 Strength",
        "+25 Dexterity",
        "+25 Intelligence",
        "+4 Magic Luck",
        "+7 Extra Equipment Slots",
    ],
    "IX" : [
        "+90% Enhanced Effect",
        "+25 Strength",
        "+25 Dexterity",
        "+25 Intelligence",
        "+4 Magic Luck",
        "+7 Extra Equipment Slots",
    ],
    "X" : [
        "+90% Enhanced Effect",
        "+30 Strength",
        "+30 Dexterity",
        "+30 Intelligence",
        "+4 Magic Luck",
        "+8 Extra Equipment Slots",
    ],
    "XI" : [
        "+90% Enhanced Effect",
        "+30 Strength",
        "+30 Dexterity",
        "+30 Intelligence",
        "+4 Magic Luck",
        "+8 Extra Equipment Slots",
    ],
    "XII" : [
        "+90% Enhanced Effect",
        "+30 Strength",
        "+30 Dexterity",
        "+30 Intelligence",
        "+4 Magic Luck",
        "+14 Life Regen",
        "+14 Mana Regen",
        "+9 Extra Equipment Slots",
    ],
    "XIII" : [
        "+90% Enhanced Effect",
        "+30 Strength",
        "+30 Dexterity",
        "+30 Intelligence",
        "+4 Magic Luck",
        "+16 Life Regen",
        "+16 Mana Regen",
        "+9 Extra Equipment Slots",
    ],
    "XIV" : [
        "+90% Enhanced Effect",
        "+30 Strength",
        "+30 Dexterity",
        "+30 Intelligence",
        "+4 Magic Luck",
        "+17 Life Regen",
        "+17 Mana Regen",
        "+10 Extra Equipment Slots",
    ]
}
prices = {
    "plain": {
        "tier" : {
            "X" : 5,
            "XI" : 6,
            "XII" : 14,
            "XIII" : 25,
            "XIV" : 60,
        },
    },
    "magical": {
        "tier" : {
            "X" : 1,
            "XI" : 1,
            "XII" : 1,
            "XIII" : 2,
            "XIV" : 10,
        },
    },
    "rare": {
        "tier" : {
            "XI" : 1,
            "XII" : 2,
            "XIII" : 2,
            "XIV" : 12,
        },
    },
    "mystical": {
        "tier" : {
            "VIII" : 1,
            "IX" : 1,
            "X" : 1,
            "XI" : 1,
            "XII" : 2,
            "XIII" : 3,
            "XIV" : 14,
        },
    },
    "angelic": {
        "tier" : {
            "VIII" : 1,
            "IX" : 1,
            "X" : 1,
            "XI" : 2,
            "XII" : 3,
            "XIII" : 5,
            "XIV" : 15,
        },
    },
    "mythical": {
        "tier" : {
            "V" : 1,
            "VI" : 1,
            "VII" : 1,
            "VIII" : 1,
            "IX" : 1,
            "X" : 2,
            "XI" : 3,
            "XII" : 4,
            "XIII" : 7,
            "XIV" : 19,
        },
    },
    "arcane": {
        "tier" : {
            "V" : 1,
            "VI" : 1,
            "VII" : 1,
            "VIII" : 2,
            "IX" : 2,
            "X" : 3,
            "XI" : 4,
            "XII" : 6,
            "XIII" : 9,
            "XIV" : 21,
        },
    },
    "legendary": {
        "tier" : {
            "V" : 1,
            "VI" : 1,
            "VII" : 1,
            "VIII" : 2,
            "IX" : 2,
            "X" : 4,
            "XI" : 5,
            "XII" : 8,
            "XIII" : 24,
            "XIV" : 35,
        },
    },
    "godly": {
        "tier" : {
            "III" : 1,
            "IV" : 1,
            "V" : 1,
            "VI" : 1,
            "VII" : 2,
            "VIII" : 3,
            "IX" : 3,
            "X" : 5,
            "XI" : 6,
            "XII" : 11,
            "XIII" : 28,
            "XIV" : 40,
        },
    },
    "epic": {
        "tier" : {
            "III" : 1,
            "IV" : 1,
            "V" : 2,
            "VI" : 2,
            "VII" : 2,
            "VIII" : 3,
            "IX" : 3,
            "X" : 6,
            "XI" : 7,
            "XII" : 12,
            "XIII" : 32,
            "XIV" : 45,
        },
    },
    "relic": {
        "tier" : {
            "III" : 1,
            "IV" : 1,
            "V" : 2,
            "VI" : 2,
            "VII" : 3,
            "VIII" : 4,
            "IX" : 4,
            "X" : 6,
            "XI" : 7,
            "XII" : 14,
            "XIII" : 36,
            "XIV" : 50,
        },
    },
    "artifact": {
        "tier" : {
            "III" : 1,
            "IV" : 1,
            "V" : 2,
            "VI" : 3,
            "VII" : 4,
            "VIII" : 5,
            "IX" : 5,
            "X" : 7,
            "XI" : 9,
            "XII" : 16,
            "XIII" : 40,
            "XIV" : 55,
        },
    },
    "unique": {
        "tier" : {
            "III" : 1,
            "IV" : 2,
            "V" : 3,
            "VI" : 4,
            "VII" : 5,
            "VIII" : 6,
            "IX" : 6,
            "X" : 8,
            "XI" : 12,
            "XII" : 19,
            "XIII" : 40,
            "XIV" : 60,
        },
    },
}


class Weapon:
    def __init__(self, weapon_popup_element, overlay_stats) -> None:
        self.name = ""
        self.rarity = ""
        self.bonus_level = overlay_stats[0]
        self.tier = overlay_stats[1]
        self.type = "weapon"
        self.level_req = 0
        self.stats = []
        self.min_damage = 0
        self.max_damage = 0
        self.parse_weapon(weapon_popup_element)
    def __repr__(self) -> str:
        return f"<Weapon {self.name}>"
    def parse_weapon(self, weapon_popup_element) -> list:
        weapon_stats = [stat.strip() for stat in weapon_popup_element.text.split('\n')]
        self.name = weapon_stats[0]
        rarities = ["magical","rare","mystical","angelic","mythical","arcane","legendary","godly","epic","relic","artifact","unique"]
        rarity = self.name.split()[0].lower()
        if rarity in rarities:
            self.rarity = rarity
        else:
            self.rarity = 'plain'
        for stat in weapon_stats[1:]:
            if "level req:" in stat.lower():
                self.level_req = int(stat.split(' ')[-1])
            if "damage:" in stat.lower():
                self.min_damage, self.max_damage = [int(num) for num in stat.split(' ') if num.isdigit()]
            if "+" in stat.lower():
                self.stats.append(stat.lower())
#-----------------------------------------------
class Armor:
    def __init__(self, armor_popup_element, overlay_stats) -> None:
        self.name = ""
        self.rarity = ""
        self.bonus_level = overlay_stats[0]
        self.tier = overlay_stats[1]
        self.type = "armor"
        self.level_req = 0
        self.stats = []
        self.min_phys_defense = 0
        self.max_phys_defense = 0
        self.min_mag_defense = 0
        self.max_mag_defense = 0
        self.parse_armor(armor_popup_element)
    def __repr__(self) -> str:
        return f"<Armor {self.name}>"
    def parse_armor(self, armor_popup_element) -> list:
        armor_stats = [stat.strip() for stat in armor_popup_element.text.split('\n')]
        self.name = armor_stats[0]
        rarities = ["magical","rare","mystical","angelic","mythical","arcane","legendary","godly","epic","relic","artifact","unique"]
        rarity = self.name.split()[0].lower()
        if rarity in rarities:
            self.rarity = rarity
        else:
            self.rarity = 'plain'
        for stat in armor_stats[1:]:
            if "level req" in stat.lower():
                self.level_req = int(stat.split(' ')[-1])
            if "physical defense" in stat.lower():
                self.min_phys_defense, self.max_phys_defense = [int(num) for num in stat.split(' ') if num.isdigit()]
            if "magic defense" in stat.lower():
                self.min_mag_defense, self.max_mag_defense = [int(num) for num in stat.split(' ') if num.isdigit()]
            if "+" in stat:
                self.stats.append(stat)
#-----------------------------------------------
class Charm:
    def __init__(self, charm_popup_element, overlay_stats) -> None:
        self.name = ""
        self.rarity = ""
        self.bonus_level = overlay_stats[0]
        self.tier = overlay_stats[1]
        self.type = "charm"
        self.level_req = 0
        self.stats = []
        self.min_spell_effect = 0
        self.max_spell_effect = 0
        self.mana_cost = 0
        self.parse_charm(charm_popup_element)
    def __repr__(self) -> str:
        return f"<Charm {self.name}>"
    def parse_charm(self, charm_popup_element) -> list:
        charm_stats = [stat.strip() for stat in charm_popup_element.text.split('\n')]
        self.name = charm_stats[0]
        rarities = ["magical","rare","mystical","angelic","mythical","arcane","legendary","godly","epic","relic","artifact","unique"]
        rarity = self.name.split()[0].lower()
        if rarity in rarities:
            self.rarity = rarity
        else:
            self.rarity = 'plain'
        for stat in charm_stats[1:]:
            if "level req" in stat.lower():
                self.level_req = int(stat.split(' ')[-1])
            if "spell damage" in stat.lower() or "heals" in stat.lower():
                self.min_spell_effect, self.max_spell_effect = [int(num) for num in stat.split(' ') if num.isdigit()]
            if "mana cost" in stat.lower():
                self.mana_cost = int(stat.split(':')[-1].strip())
            if "+" in stat.lower():
                self.stats.append(stat.lower())
#-----------------------------------------------



import pdb
import logging
import platform
import undetected_chromedriver as uc
from ladderbot.gui import show_gui
from ladderbot import controllers
import importlib.resources
#-----------------------------------------------
def bot():
    logging.basicConfig(
        level=logging.INFO,
        format='%(asctime)s - %(levelname)s - %(message)s',
        filename=str(importlib.resources.files("ladderbot")) + "\\debug.log",
        filemode='w'
    )
    # Create logger
    logger = logging.getLogger()
    show_gui()
    # Set the options for the Chrome driver
    chrome_options = uc.ChromeOptions()
    prefs = {
        "credentials_enable_service": False,
        "profile.password_manager_enabled": False
    }
    # Set the window position
    chrome_options.add_argument("--window-position=-5,695")
    chrome_options.add_experimental_option("prefs", prefs)
    # Check the system's operating system
    platform_name = platform.system()
    logger.info(f"Running on {platform_name}")
    driver = uc.Chrome(options=chrome_options)
    # Set the window size
    driver.set_window_size(900, 705)
    # Initialize all controllers before interacting with the driver or driver de-sync can occur
    login_controller = controllers.LoginController(driver, logger)
    player_controller = controllers.PlayerController(driver, logger)
    leveling_controller = controllers.LevelingController(driver, logger)
    inventory_controller = controllers.InventoryController(driver, logger)
    transmute_controller = controllers.TransmuteController(driver, logger)
    market = controllers.MarketController(driver, logger, gold_password='temp')
    vault = controllers.VaultController(driver, logger)
    # Create Login instance and boot the game
    login = controllers.Login(login_controller)
    login.run(character_name='kdasje5465131')
    # Create Player and Inventory instances to maintain game data
    inventory = controllers.Inventory(inventory_controller)
    player = controllers.Player(logger, player_controller, inventory)
    while True:
            # try:
                leveling_controller.run()
                player.update_health()
                player.update_transmute_rank()
                player.inventory.load()
                player.do_master_quest()
                pdb.set_trace()
                # transmute_controller.transmute_equipment(player.transmute_rank, player.inventory.equipment)
                # vault.deposit_equipment(player.inventory.equipment)
                # market.sell_equipment(player.inventory.equipment)
            # except:
            #     driver.close()
            #     return
#-----------------------------------------------


# import os
import importlib.resources
from time import sleep
from random import choice
from pathlib import Path
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import Select
from selenium.webdriver import Keys, ActionChains
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait
from selenium.common.exceptions import NoSuchElementException, StaleElementReferenceException, TimeoutException
from ladderbot.items import Weapon, Armor, Charm
from ladderbot.mappings import transmute_mapping, item_pickup_mapping, deposit_criteria, prices
#-----------------------------------------------
class LoginController:
    def __init__(self, driver, logger) -> None:
        self.name = "LoginController"
        self.logger = logger
        # self.username = username
        # self.password = password
        self.driver = driver
    # def login(self):
    #     self.driver.get('https://forums.d2jsp.org/login.php')
    #     # Enter login credentials and submit the form
    #     username_field = WebDriverWait(self.driver, 10).until(
    #         EC.presence_of_element_located((By.NAME, 'username'))
    #     )
    #     # username_field = self.driver.find_element(By.NAME, 'username')
    #     username_field.send_keys(self.username)  # Replace YOUR_USERNAME with your actual username
    #     username_field.send_keys(Keys.RETURN)

    #     # password_field = self.driver.find_element(By.NAME, 'pass')
    #     password_field = WebDriverWait(self.driver, 10).until(
    #         EC.visibility_of_element_located((By.NAME, 'pass'))
    #     )
    #     sleep(.1)
    #     password_field.send_keys(self.password)  # Replace YOUR_PASSWORD with your actual password
    #     sleep(10)
    #     password_field.send_keys(Keys.RETURN)
    #     sleep(2)
    #     return self.driver.current_url == 'https://forums.d2jsp.org/'
    # def logout(self):
    #     self.driver.get("https://forums.d2jsp.org/login.php?c=8")
    #     logout_button = WebDriverWait(self.driver, 5).until(
    #         EC.presence_of_element_located((By.XPATH, '/html/body/form/dl/dd/div[3]/input'))
    #     )
    #     logout_button.click()
    def check_for_element(self, class_name):
        status = False
        try:
            # check if game is started
            # self.driver.find_element(By.CLASS_NAME, class_name)
            WebDriverWait(self.driver, 2).until(
                EC.presence_of_element_located((By.CLASS_NAME, class_name))
            )
            status = True
        except:
            # element wasnt found
            pass
        return status
    def check_location(self) -> bool:
        self.logger.info(f"{self.name}.check_location() --> Reading player location")
        indicators = {
            'town' : (By.CLASS_NAME, 'townOIcon'),
            'catacombs' : (By.CLASS_NAME, 'bgCata'),
            'pond' : (By.XPATH, '//*[@id="bg"]/img'),
            'character_selection' : (By.XPATH, '//*[@id="bg"]/div[1]/a'),
            'character_selected' : (By.XPATH, '//*[@id="bg"]/div[2]/div/a[2]'),
            'character_creation' : (By.XPATH, '//*[@id="bg"]/div/div[2]/div[1]'),
        }
        for location in indicators:
            try:
                WebDriverWait(self.driver, 1).until(
                    EC.presence_of_element_located(indicators.get(location))
                )
                self.logger.info(f"{self.name}.check_location() --> {location=}")
                return location
            except Exception as e:
                pass
        else:
            self.logger.error(f"{self.name}.check_location() --> Location Unknown!")
            return False
    def leave_cata(self):
        self.logger.info(f"{self.name}.leave_cata() --> Leaving Catacombs")
        # Spams exiting the catacombs until we've left
        while self.check_location() == 'catacombs':
            try:
                to_town_button = self.driver.find_element(By.CLASS_NAME, "gradRed")
                to_town_button.click()
            except Exception as e:
                # Failed to leave catacombs, keep trying forever. or die...
                pass
    def start_game(self):
        html_file = Path(str(importlib.resources.files("ladderbot")) + "\\index.html")
        play_url = html_file.as_uri()
        self.logger.info(f"{self.name}.start_game() --> Starting Game {play_url=}")
        self.driver.get(play_url)
        try: 
            play_button = WebDriverWait(self.driver, 5).until(
                EC.presence_of_element_located((By.CLASS_NAME, 'abutGradBl'))
            )
            play_button.click()
        except TimeoutError as e:
            raise e
        try: 
            stage = WebDriverWait(self.driver, 5).until(
                EC.presence_of_element_located((By.ID, 'stage'))
            )
            stage.click()
        except TimeoutError as e:
            raise e
    def select_character(self, name:str):
        self.logger.info(f"{self.name}.select_character() --> {name=}")
        character_list = WebDriverWait(self.driver, 5).until(
            EC.presence_of_element_located((By.CLASS_NAME, 'charList'))
        )
        available_characters = character_list.find_elements(By.CLASS_NAME, 'cName')
        character = list(filter(lambda x:x.text == name, available_characters))[0]
        character.click()
        character_play_button = self.driver.find_element(By.CLASS_NAME, 'clPlay')
        character_play_button.click()
    def logout_character(self):
        self.logger.info(f"{self.name}.logout_character()")
        button_grouping = self.driver.find_element(By.CLASS_NAME, 'ctrlButtons')
        logout_character_button = button_grouping.find_elements(By.TAG_NAME, 'img')[-1]
        logout_character_button.click()
class Login:
    def __init__(self, controller:LoginController) -> None:
        self.controller = controller
    def run(self, character_name:str):
        # self.controller.login()
        # launch the game
        self.controller.start_game()
        current_location = self.controller.check_location()
        if current_location == 'catacombs':
            self.controller.leave_cata()
            self.controller.logout_character()
            self.controller.select_character(character_name)
        # already at the town
        elif current_location == 'town':
            # logout current character
            self.controller.logout_character()
            self.controller.select_character(character_name)
        # login desired character
        elif current_location == 'character_selection':
            self.controller.select_character(character_name)
#-----------------------------------------------
class InventoryController:
    def __init__(self, driver, logger) -> None:
        self.name = "InventoryController"
        self.driver = driver
        self.logger = logger
        self.action = ActionChains(self.driver)
        self.weapon = None
        self.armor = None
        self.charm = None
        self.acc_charm = None
    def open(self):
        self.action.send_keys(Keys.SPACE).perform()
        self.action.send_keys("i").perform()
        try:
            WebDriverWait(self.driver, 3).until(
                EC.visibility_of_element_located((By.CLASS_NAME, 'pbInv'))
            )
        except TimeoutError as e:
            raise e
    def close_all(self):
        self.action.send_keys(Keys.SPACE).perform()
    def get_equipment_count(self):
        self.logger.info(f"{self.name} --> getting inventory capacity")
        self.open()
        equipment_label_element = WebDriverWait(self.driver, 1).until(
            EC.visibility_of_element_located((By.CLASS_NAME, 'invEqCountLabel'))
        )
        label_text = equipment_label_element.text
        current, capacity = [int(letter) for letter in label_text.split() if letter.isdigit()]
        self.logger.info(f"{self.name} --> {current=}, {capacity=}")
        self.close_all()
        return current, capacity
    def load_equipped_items(self):
        self.logger.info(f"{self.name} --> reading equipped items from inventory")
        self.open()
        equipped_items_container = WebDriverWait(self.driver, .5).until(
            EC.visibility_of_element_located((By.CLASS_NAME, 'invEquipped'))
        ) 
        equipped_items = WebDriverWait(equipped_items_container, .5).until(
            EC.visibility_of_all_elements_located((By.CLASS_NAME, 'itemSlotBox'))
        )
        weapon, armor, charm, acc_charm = equipped_items
        overlay_stats = weapon.text.replace('+',"").split('\n')
        if len(overlay_stats) == 1:
                    overlay_stats.insert(0, '0')
        self.action.move_to_element(weapon).perform()
        try:
            weapon_stats_popup = WebDriverWait(self.driver, .5).until(
                EC.visibility_of_element_located((By.CLASS_NAME, 'tbItemDesc'))
            )
            weapon = Weapon(weapon_stats_popup, overlay_stats)
        except TimeoutException:
            self.logger.info(f"{self.name} --> TimeoutException: weapon not equipped")
            weapon = None
        overlay_stats = armor.text.replace('+',"").split('\n')
        if len(overlay_stats) == 1:
                    overlay_stats.insert(0, '0')
        self.action.move_to_element(armor).perform()
        try:
            armor_stats_popup = WebDriverWait(self.driver, .5).until(
                EC.visibility_of_element_located((By.CLASS_NAME, 'tbItemDesc'))
            )
            armor = Armor(armor_stats_popup, overlay_stats)
        except TimeoutException:
            self.logger.info(f"{self.name} --> TimeoutException: armor not equipped")
            armor = None
        overlay_stats = charm.text.replace('+',"").split('\n')
        if len(overlay_stats) == 1:
                    overlay_stats.insert(0, '0')
        self.action.move_to_element(charm).perform()
        try:
            charm_stats_popup = WebDriverWait(self.driver, 1).until(
                EC.visibility_of_element_located((By.CLASS_NAME, 'tbItemDesc'))
            )
            charm = Charm(charm_stats_popup, overlay_stats)
        except TimeoutException:
            self.logger.info(f"{self.name} --> TimeoutException: charm not equipped")
            charm = None
        overlay_stats = acc_charm.text.replace('+',"").split('\n')
        if len(overlay_stats) == 1:
                    overlay_stats.insert(0, '0')
        self.action.move_to_element(acc_charm).perform()
        try:
            charm_stats_popup = WebDriverWait(self.driver, 1).until(
                EC.visibility_of_element_located((By.CLASS_NAME, 'tbItemDesc'))
            )
            acc_charm = Charm(charm_stats_popup, overlay_stats)
        except TimeoutException:
            self.logger.info(f"{self.name} --> TimeoutException: acc_charm not equipped")
            acc_charm = None
        return weapon, armor, charm, acc_charm
    def load_equipment(self):
        self.logger.info(f"{self.name} --> reading inventory equipment")
        equipment = {}
        self.action.send_keys(Keys.SPACE).perform()
        self.action.send_keys("i").perform()
        equipment_container = WebDriverWait(self.driver, 1).until(
            EC.visibility_of_element_located((By.CLASS_NAME, 'invEqBox'))
        )
        equipment_elements = WebDriverWait(equipment_container, 1).until(
            EC.visibility_of_all_elements_located((By.CLASS_NAME, 'itemSlotBox'))
        )
        # Enumerate the equipment slots to see if there is an item,
        #   and if so in which slot
        for index, element in enumerate(equipment_elements):
            self.logger.info(f"{self.name} --> reading inventory item {index} icon tier/bonus overlay")
            try:
                item_element = element.find_element(By.CLASS_NAME, 'itemBox')
            except NoSuchElementException:
                self.logger.info(f"{self.name} --> NoSuchElementException: No equipment in slot {index}")
                equipment[index] = None
                continue
            overlay_stats = item_element.text.replace('+',"").split('\n')
            if len(overlay_stats) == 1:
                overlay_stats.insert(0, '0')
            self.action.move_to_element(item_element).perform()
            item_comparison_grouping = WebDriverWait(self.driver, 1).until(
                EC.visibility_of_element_located((By.CLASS_NAME, 'tbItemDesc'))
            )
            item_description_element = item_comparison_grouping.find_elements(By.XPATH, "./child::*")[0]
            item = self.convert_item_from_element(item_description_element, overlay_stats)
            equipment[index] = item
        return equipment
    def classify_item(self, item_name:str) -> str:
        self.logger.info(f"{self.name} --> classifying {item_name=}")
        weapon_types = ["sword","club","axe","dagger","staff","longsword","warhammer","battleaxe","spear","polearm"]
        charm_types = ["ice","fire","lightning","wind","earth","wild Heal","heal","focused heal"]
        armor_types = ["robe","padded Robe","leather armor","scale armor","chain mail","plate mail"]
        for weapon_type in weapon_types:
            if weapon_type in item_name.lower():
                self.logger.info(f"{self.name} --> {item_name=} is weapon")
                return "weapon"
        for armor_type in armor_types:
            if armor_type in item_name.lower():
                self.logger.info(f"{self.name} --> {item_name=} is armor")
                return "armor"
        for charm_type in charm_types:
            if charm_type in item_name.lower():
                self.logger.info(f"{self.name} --> {item_name=} is charm")
                return "charm"         
    def convert_item_from_element(self, item_popup_element, overlay_stats):
        item_name = item_popup_element.text.split("\n")[0]
        self.logger.info(f"{self.name} --> parsing {item_name=} into item object")
        item_classification = self.classify_item(item_name)
        if item_classification == "weapon":
            item = Weapon(item_popup_element, overlay_stats)
            self.logger.info(f"{self.name} --> {item_name=} parsed to weapon item object")
        elif item_classification == "armor":
            item = Armor(item_popup_element, overlay_stats)
            self.logger.info(f"{self.name} --> {item_name=} parsed to armor item object")
        elif item_classification == "charm":
            item = Charm(item_popup_element, overlay_stats)
            self.logger.info(f"{self.name} --> {item_name=} parsed to charm item object")
        else:
            self.logger.error(f"{self.name} --> could not parse {item_name=} to item object")
            item = None
        return item        
    def delete_item(self, item_element):
        inventory_window_element = WebDriverWait(self.driver, 3).until(
            EC.presence_of_element_located((By.CLASS_NAME, 'invIWSide'))
        )
        trash_bin_element = inventory_window_element.find_element(By.XPATH, '//*[@id="stage"]/div[5]/div[4]/div[3]/div')
        self.action.drag_and_drop(item_element, trash_bin_element).perform()
class Inventory:
    def __init__(self, inventory_controller:InventoryController) -> None:
        self.controller = inventory_controller
        self.weapon = None
        self.armor = None
        self.charm = None
        self.acc_charm = None
        self.equipment = None
    def load(self):
        self.weapon, self.armor, self.charm, self.acc_charm = self.controller.load_equipped_items()
        self.equipment = self.controller.load_equipment()
        self.controller.close_all()
    def is_full(self):
        return self.controller.get_equipment_count()[0] >= 9
#-----------------------------------------------
class PlayerController:
    def __init__(self, driver, logger) -> None:
        self.name = "PlayerController"
        self.driver = driver
        self.logger = logger
        self.action = ActionChains(self.driver)
        self.location_map = {
            'marketplace':0,
            'catacombs':1,
            'shrine':2,
            'vault':3,
            'pond':4,
            'cooking':5,
            'transmuting':6,
            'glyphing':7,
            'suffusencing':8,
            'master_quest':9
        }
    def check_health(self):
        # health_element = self.driver.find_element(By.CLASS_NAME, 'lifeMeter')
        health_element = WebDriverWait(self.driver, 3).until(
            EC.presence_of_element_located((By.CLASS_NAME, 'lifeMeter'))
        )
        current_health, max_health = health_element.text.replace(' ', '').split('/')
        return {
            'current' : int(current_health),
            'max_health' : int(max_health),
            'percent' : int(100*(int(current_health)/int(max_health)))
        }
    def check_location(self) -> bool:
        indicators = {
            'town' : (By.CLASS_NAME, 'townOIcon'),
            'catacombs' : (By.CLASS_NAME, 'bgCata'),
            'pond' : (By.CLASS_NAME, 'fishDock'),
            'master_quest' : (By.CLASS_NAME, 'mqDoor'),
            'character_creation' : (By.CLASS_NAME, 'createCharWrap'),
            'character_selected' : (By.CLASS_NAME, 'clCharWrap'),
            'character_selection' : (By.XPATH, '//*[@id="bg"]/div[1]/a'),
        }
        for location in indicators:
            try:
                self.driver.find_element(*(indicators.get(location)))
                self.logger.info(f"{self.name}.check_location() --> at {location}")
                return location
            except Exception as e:
                # If an exception occured check the next location indicator
                self.logger.info(f"{self.name}.check_location() --> not at {location}")
                pass
        else:
            # print(f'could not determine location')
            return False
    def check_mana(self):
        mana_element = self.driver.find_element(By.CLASS_NAME, 'manaMeter')
        current_mana, max_mana = mana_element.text.replace(' ', '').split('/')
        return {
            'current' : int(current_mana),
            'max_mana' : int(max_mana),
            'percent' : int(100*(int(current_mana)/int(max_mana)))
        }
    def go(self, location:str):
        # only expected to be used while player is at the Town
        self.logger.info(f"{self.name}.go() --> traveling {location=}")
        while True:
            self.close_all()
            location_button = WebDriverWait(self.driver, 3).until(
                EC.visibility_of_element_located((By.ID, f'townObj{self.location_map[location]}'))
            )
            location_button.click()
            current_location = self.check_location()
            sleep(.2)
            if current_location != location:
                self.logger.info(f"{self.name}.go() --> error traveling {location=}")
                sleep(8)
            else:
                return location
  
    def close_all(self):
        ActionChains(self.driver)\
        .send_keys(Keys.SPACE)\
        .perform()
    def move(self, direction:str):
        if direction == 'forward':
            ActionChains(self.driver)\
            .send_keys(Keys.UP)\
            .perform()
            return
        elif direction == 'backward':
            ActionChains(self.driver)\
            .send_keys(Keys.DOWN)\
            .perform()
            return
        elif direction == 'left':
            ActionChains(self.driver)\
            .send_keys(Keys.LEFT)\
            .perform()
            return
        elif direction == 'right':
            ActionChains(self.driver)\
            .send_keys(Keys.RIGHT)\
            .perform()
            return
        elif direction == 'whistle':
            ActionChains(self.driver)\
            .send_keys('g')\
            .perform()
    def exit_catacombs(self):
        # Spams exiting the catacombs until we've left
        while self.check_location() == 'catacombs':
            try:
                to_town_button =self.driver.find_element(By.CLASS_NAME, "gradRed")
                to_town_button.click()
            except Exception as e:
                # Failed to leave catacombs, keep trying forever. or die...
                pass
    def mobs_on_screen(self):
        mobs =  self.driver.find_elements(By.CLASS_NAME, "mob")
        return mobs
    def use_abilities(self):
        try:
            ability_elements = WebDriverWait(self.driver, 1).until(
                EC.visibility_of_all_elements_located((By.CLASS_NAME, 'bbAbility'))
            )
        except:
            return
        try:
            ability_elements.reverse()
            for element in ability_elements:
                if 'gradRed' in element.get_attribute('class').split():
                    return
                if int(element.text.split()[0].replace('%', '')) >= 50:
                    element.click()
                    return
        except StaleElementReferenceException as e:
            # an ability was used and disappeared unexpectedly
            return
        except TypeError as e:
            # no ability elements were found
            return
    def attack_mob(self):
        ActionChains(self.driver)\
            .send_keys('q')\
            .perform()
        try:
            mob = self.driver.find_elements(By.CLASS_NAME, "mob")[0]
            mob.click()
        except Exception as e:
            # print('Mob died before it could be clicked')
            self.move("whistle")      
    def check_for_drops(self):
        drop_elements = None
        try:
            drops_container = self.driver.find_element(By.CLASS_NAME, "dropItemsBox")
            drop_elements = drops_container.find_elements(By.CLASS_NAME, "itemBox")
        except:
            # if finding this container errors out, there's no drops
            pass
        return drop_elements
    def _should_pickup_drop(self, item):
        self.action.send_keys("p").perform()
        proficiencies_container_element = WebDriverWait(self.driver, 3).until(
            EC.visibility_of_element_located((By.CLASS_NAME, 'pbProfs'))
        )
        transmute_proficiency_element = WebDriverWait(proficiencies_container_element, 3).until(
            EC.visibility_of_all_elements_located((By.CLASS_NAME, 'profStat'))
        )[-2]
        player_transmute_rank = int(transmute_proficiency_element.text)
        self.action.send_keys("p").perform()
        tier_mapping = item_pickup_mapping.get(item.tier)
        if item.rarity.lower() in tier_mapping['rarity_levels']:
            if tier_mapping["ranks"]["min"] <= player_transmute_rank <= tier_mapping["ranks"]["max"]:
                return True
        return False
    def should_pickup_drop(self, item):
        if item.type == "weapon" and item.max_damage > self.inventory.weapon.max_damage:
            # print(f"Found {item}")
            return True
        elif item.type == "armor" and item.max_phys_defense > self.inventory.armor.max_phys_defense:
            # print(f"Found {item}")
            return True
        elif item.type == "charm" and item.max_spell_effect > self.inventory.charm.max_spell_effect:
            # print(f"Found {item}")
            return True
        else:
            # print(f"Ignoring: {item}")
            pass
        return False
class Player:
    def __init__(self, logger, player_controller:PlayerController, inventory:Inventory) -> None:
        self.name = "Player"
        self.logger = logger
        self.controller: PlayerController = player_controller
        self.health: dict = self.controller.check_health()
        self.mana: dict = self.controller.check_mana()
        self.location: str = self.controller.check_location()
        self.inventory: Inventory = inventory
        if self.location == "catacombs":
            self.exit_catacombs()
        self.transmute_rank: int = self.update_transmute_rank()
        # self.inventory.load()
    def enter_catacombs(self):
        self.controller.go('catacombs')
    def exit_catacombs(self):
        self.logger.info(f"{self.name}.exit_catacombs() --> Leaving Catacombs")
        self.controller.exit_catacombs()
    def update_health(self):
        self.health = self.controller.check_health()
    def update_transmute_rank(self):
        self.controller.action.send_keys(Keys.SPACE).perform()
        self.controller.action.send_keys("p").perform()
        proficiencies_container_element = WebDriverWait(self.controller.driver, 3).until(
            EC.visibility_of_element_located((By.CLASS_NAME, 'pbProfs'))
        )
        transmute_proficiency_element = WebDriverWait(proficiencies_container_element, 3).until(
            EC.visibility_of_all_elements_located((By.CLASS_NAME, 'profStat'))
        )[-2]
        self.transmute_rank = int(transmute_proficiency_element.text)
        self.controller.action.send_keys(Keys.SPACE).perform()
    def explore(self, exit_health_percent):
        retreat = False
        self.controller.go('catacombs')
        while retreat == False:
            combat_occurred = False
            movement_direction = choice(['forward', 'left', 'right'])
            self.controller.move(movement_direction)
            # Continiously look to see if there's a mob on screen
            while True:
                self.update_health()
                # Pick up drops if there are any and get True/False result
                dropped_item_elements = self.controller.check_for_drops()
                if dropped_item_elements:
                    # work the list from back to front to preserve element order
                    dropped_item_elements.reverse()
                    for item_element in dropped_item_elements:
                        try:
                            self.inventory.controller.action.move_to_element(item_element).perform()
                        except StaleElementReferenceException:
                            self.logger.info(f"{self.name} --> StaleElementReferenceException: Element in dropped items container became stale")
                            continue
                        try:
                            item_comparison_grouping = self.controller.driver.find_element(By.CLASS_NAME, 'tbItemDesc')
                        except NoSuchElementException:
                            self.logger.info(f"{self.name} --> NoSuchElementException: item_comparison_grouping could not be located")
                            continue
                        try:
                            item_description_element = item_comparison_grouping.find_elements(By.XPATH, "./child::*")[0]
                        except StaleElementReferenceException:
                            self.logger.info(f"{self.name} --> StaleElementReferenceException: item_comparison_grouping stale")
                            continue
                        overlay_stats = item_element.text.replace('+',"").split('\n')
                        if len(overlay_stats) == 1:
                            overlay_stats.insert(0, '0')
                        item = self.inventory.controller.convert_item_from_element(item_description_element, overlay_stats)
                        if self.controller._should_pickup_drop(item):
                            item_element.click()
                # If we picked something up the inventory might've become full
                if dropped_item_elements and self.inventory.is_full():
                    retreat = True
                    break
                if self.health['percent'] <= exit_health_percent:
                    retreat = True
                    break
                # if there are 9 mobs on screen likely they replicate too fast to be killed
                # joining and leaving a group will reset the catacombs and mobs
                active_mobs = self.controller.mobs_on_screen()
                if len(active_mobs) == 9:
                    self.logger.info(f"{self.name}.explore() --> Too many mobs to combat")
                    # command to join a goup
                    self.logger.info(f"{self.name}.explore() --> Joining group")
                    command = "njs.sendBytes(60, 3, 2, 5)"
                    self.controller.driver.execute_script(command)
                    sleep(2)
                    self.logger.info(f"{self.name}.explore() --> Leaving group")
                    # command to leave a group
                    command = "njs.sendBytes(60, 5)"
                    self.controller.driver.execute_script(command)
                    break
                # See if there are any mobs attacking the player
                if 0 < len(active_mobs) < 9:
                    mob_names = []
                    for name in active_mobs:
                        try:
                            mob_names.append(name.text)
                        except:
                            pass
                    # If this is the first time we've noticed the group attacking us
                    if combat_occurred == False:
                        # Show which mobs we are in battle with
                        # print(f"\nIn battle with: {mob_names}")
                        combat_occurred = True
                    # use an ability if it is ready
                    self.controller.use_abilities()
                    # and attack them!
                    self.controller.attack_mob()
                
                else:
                    break
            # If the player saw combat
            if combat_occurred == True:
                # Get how much health is remaining
                self.health = self.controller.check_health()
                # then to re-engage with the maze
                self.controller.move("whistle")
        self.exit_catacombs()
    def rest(self, stop_resting_health=100):
        while self.health['percent'] < stop_resting_health:
            self.update_health()
            sleep(.25)
    def do_master_quest(self):
        self.controller.go("master_quest")
#-----------------------------------------------
class MarketController:
    def __init__(self, driver, logger, gold_password) -> None:
        self.name = "MarketController"
        self.logger = logger
        self.gold_password = gold_password
        self.driver = driver
        self.action = ActionChains(self.driver)  
    def open(self):
        self.action.send_keys(Keys.SPACE).perform()
        self.action.send_keys("m").perform()
        try:
            WebDriverWait(self.driver, 3).until(
                EC.visibility_of_element_located((By.CLASS_NAME, 'marketTabs'))
            )
        except TimeoutError as e:
            raise e
    def _open_sell_tab(self):
        self.open()
        market_tabs = WebDriverWait(self.driver, 3).until(
            EC.visibility_of_element_located((By.CLASS_NAME, 'marketTabs'))
        )
        search, sell, transfer = market_tabs.find_elements(By.CLASS_NAME, 'njRB')
        sell.click()
        return
    def search(self, type:str=None, subtype:str=None, min_level:int=None, max_level:int=None, min_magic:int=None, max_magic:int=None, min_cost:int=0, max_cost:int=999999, attribute1:str=None, attribute2:str=None, attribute3:str=None, only_equipable:bool=False):
        market_controls = self.driver.find_element(By.CLASS_NAME, 'marketControls')
        selection_fields = Select(market_controls.find_elements(By.TAG_NAME, 'select'))
        if type:
            type_select = selection_fields[0]
            type_select.select_by_visible_text(type)
        if subtype:
            subtype_select = selection_fields[1]
            subtype_select.select_by_visible_text(subtype)
        level_map = {
                1: "I (0)",
                2: "II (0)",
                3: "III (0)",
                5: "IV (5)",
                10: "V (10)",
                15: "VI (15)",
                20: "VII (20)",
                25: "VIII (25)",
                30: "IX (30)",
                35: "X (35)",
                40: "XI (40)",
                45: "XII (45)",
                50: "XIII (50)",
                55: "XIV (55)"
            }        
        if min_level:
            min_level_select = selection_fields[2]
            min_level_select.select_by_visible_text(level_map[min_level])
        if max_level:
            max_level_select = selection_fields[3]
            max_level_select.select_by_visible_text(level_map[max_level])
        if min_magic:
            min_magic_select = selection_fields[4]
            min_magic_select.select_by_visible_text(level_map[min_magic])
        if max_magic:
            max_magic_select = selection_fields[5]
            max_magic_select.select_by_visible_text(level_map[max_magic])
        if min_cost:
            self.driver.find_element(By.XPATH, '//*[@id="stage"]/div[5]/div[2]/div[1]/div[5]/input[1]').send_keys(str(min_cost))
        if max_cost and max_cost != 999999:
            self.driver.find_element(By.XPATH, '//*[@id="stage"]/div[5]/div[2]/div[1]/div[5]/input[2]').send_keys(str(max_cost))
        if attribute1:
            attribute1 = {
                'Enhanced Effect' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[2]',
                'Strength' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[3]',
                'Dexterity' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[4]',
                'Vitality' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[5]',
                'Intelligence' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[6]',
                'Max Life' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[7]',
                'Max Mana' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[2]',
                'Experience Gained' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[8]',
                'Magic Luck' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[9]',
                'Life Regen' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[10]',
                'Mana Regen' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[11]',
                'Extra Equipment Slots' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[12]',
                'Critical Strike' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[13]',
                'Life per Attack' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[14]',
                'Mana per Attack' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[15]',
                'Life per Kill' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[16]',
                'Mana per Kill' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[17]',
                'Life Steal' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[18]',
                'Damage Return' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[19]',
                'Mind Numb' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[20]',
                'Armor Pierce' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[22]',
                'Parry' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[22]',
                'Critical Flux' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[22]',
                'Physical Damage Reduction' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[24]',
                'Magical Damage Reduction' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[25]',
                'Mana Syphon' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[26]',
                'Quick Draw' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[27]',
                'Mana Consumption' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[28]',
                'Heal Mastery' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[29]',
                'Mana Skin' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[30]',
                'Power Shot' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[31]',
                'Glancing Blow' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[32]',
                'Jubilance' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[33]',
                'Ice Mastery' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[34]',
                'Fire Mastery' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[35]',
                'Lightning Mastery' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[36]',
                'Wind Mastery' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[37]',
                'Earth Mastery' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[38]',
                'Quantity' : '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/select/option[39]'
            }
            self.driver.find_element(By.XPATH, '//*[@id="stage"]/div[5]/div[2]/div[1]/div[6]/input').send_keys(str(1))
            self.driver.find_element(By.XPATH, attribute1['Vitality']).click()
        if attribute2:
            attribute2 = {k:v.replace('div[6]','div[7]') for k,v in attribute1.items()}
            self.driver.find_element(By.XPATH, '//*[@id="stage"]/div[5]/div[2]/div[1]/div[7]/input').send_keys(str(1))
            self.driver.find_element(By.XPATH, attribute2['Vitality']).click()
        if attribute3:
            attribute3 = {k:v.replace('div[6]','div[8]') for k,v in attribute1.items()}
            self.driver.find_element(By.XPATH, '//*[@id="stage"]/div[5]/div[2]/div[1]/div[8]/input').send_keys(str(1))
            self.driver.find_element(By.XPATH, attribute3['Vitality']).click()

    def sell_equipment(self, player_equipment):
        for item_index in player_equipment:
            if player_equipment[item_index] is None:
                self.logger.info(f"{self.name} --> No equipment in slot {item_index}")
                continue
            rarity = player_equipment[item_index].rarity
            tier = player_equipment[item_index].tier
            price = self._should_sell(rarity, tier)
            if price is not None:
                self._sell(item_index, price)
        self.action.send_keys(Keys.SPACE).perform()
    def _sell(self, equipment_item_index, price):
        self.action.send_keys(Keys.SPACE).perform()
        self.action.send_keys("m").perform()
        self._open_sell_tab()
        equipment_container = WebDriverWait(self.driver, 3).until(
            EC.visibility_of_element_located((By.CLASS_NAME, 'invEqBox'))
        )
        equipment_elements = WebDriverWait(equipment_container, 3).until(
            EC.visibility_of_all_elements_located((By.CLASS_NAME, 'itemSlotBox'))
        )
        equipment_element_to_sell = equipment_elements[equipment_item_index]
        ActionChains(self.driver) \
        .key_down(Keys.SHIFT) \
        .click(equipment_element_to_sell) \
        .key_up(Keys.SHIFT) \
        .perform()
        cost_field = WebDriverWait(self.driver, 3).until(
            EC.presence_of_element_located((By.ID, 'mkSellCost'))
        )
        cost_field.clear()
        cost_field.send_keys(f"{price}")  # Replace YOUR_USERNAME with your actual username
        gold_pw_field = WebDriverWait(self.driver, 3).until(
            EC.presence_of_element_located((By.ID, 'mkGpwd'))
        )
        gold_pw_field.clear()
        gold_pw_field.send_keys(self.gold_password)
        sell_item_button = WebDriverWait(self.driver, 3).until(
            EC.presence_of_element_located((By.ID, 'mkBuySell'))
        )
        sell_item_button.click()
        sleep(1)
    
    def _should_sell(self, item_rarity, item_tier):
        prices_per_level = prices.get(item_rarity)
        if prices_per_level == None:
            return False
        price = prices_per_level.get(item_tier)
        # print(f"{price=}")
        return price
#-----------------------------------------------
class VaultController:
    def __init__(self, driver, logger) -> None:
        self.name = "VaultController"
        self.logger = logger
        self.driver = driver
        self.action = ActionChains(self.driver)
    def open(self):
        self.action.send_keys(Keys.SPACE).perform()
        self.action.send_keys("v").perform()
        try:
            WebDriverWait(self.driver, 3).until(
                EC.visibility_of_element_located((By.CLASS_NAME, 'vaultTabs'))
            )
        except TimeoutError as e:
            raise e
    def _open_deposit_tab(self):
        self.open()
        market_tabs = WebDriverWait(self.driver, 3).until(
            EC.visibility_of_element_located((By.CLASS_NAME, 'vaultTabs'))
        )
        vault, withdraw, deposit, transfer = market_tabs.find_elements(By.CLASS_NAME, 'njRB')
        deposit.click()
        return
    def deposit_equipment(self, player_equipment):
        self.action.send_keys(Keys.SPACE).perform()
        self._open_deposit_tab()
        for item_index in player_equipment:
            if player_equipment[item_index] is None:
                self.logger.info(f"{self.name} --> no equipment in slot {item_index}")
                continue
            stats = player_equipment[item_index].stats
            tier = player_equipment[item_index].tier
            if self._should_deposit(tier, stats):
                self._deposit(item_index)
        self.action.send_keys(Keys.SPACE).perform()
    def _deposit(self, equipment_item_index):
        equipment_container = WebDriverWait(self.driver, 3).until(
            EC.visibility_of_element_located((By.CLASS_NAME, 'invEqBox'))
        )
        equipment_elements = WebDriverWait(equipment_container, 3).until(
            EC.visibility_of_all_elements_located((By.CLASS_NAME, 'itemSlotBox'))
        )
        equipment_element_to_deposit = equipment_elements[equipment_item_index]
        ActionChains(self.driver) \
        .key_down(Keys.SHIFT) \
        .click(equipment_element_to_deposit) \
        .key_up(Keys.SHIFT) \
        .perform()
        sell_item_button = WebDriverWait(self.driver, 3).until(
            EC.presence_of_element_located((By.CLASS_NAME, 'vDepB'))
        )
        sell_item_button.click()
        sleep(1)
    def _should_deposit(self, item_tier, item_stats):
        stats_per_tier = deposit_criteria.get(item_tier)
        if stats_per_tier == None:
            return False
        for stat in item_stats:
            if stat in stats_per_tier:
                return True
        return False
#-----------------------------------------------
class TransmuteController:
    def __init__(self, driver, logger) -> None:
        self.name = "TransmuteController"
        self.logger = logger
        self.driver = driver
        self.action = ActionChains(self.driver)
    def transmute_equipment(self, player_transmute_rank, player_equipment):
        self.logger.info(f"{self.name} --> Starting transmute routine")
        self.action.send_keys(Keys.SPACE).perform()
        for item_index in player_equipment:
            item = player_equipment[item_index]
            if item and self._should_transmute(item, player_transmute_rank):
                self._transmute(item_index)
    def _should_transmute(self, item, player_transmute_rank):
        self.logger.info(f"{self.name} --> checking if {item} should be transmuted")
        tier_mapping = transmute_mapping.get(item.tier)
        if item.rarity.lower() in tier_mapping['rarity_levels']:
            if tier_mapping["ranks"]["min"] <= player_transmute_rank <= tier_mapping["ranks"]["max"]:
                return True
        return False
    def _transmute(self, equipment_item_index):
        self.logger.info(f"{self.name} --> transmuting item {equipment_item_index}")
        self.action.send_keys(Keys.SPACE).perform()
        # open transmute menu
        transmute_town_element = WebDriverWait(self.driver, 3).until(
            EC.visibility_of_element_located((By.ID, 'townObj6'))
        )
        transmute_town_element.click()
        # get equipment container
        equipment_container = WebDriverWait(self.driver, 3).until(
            EC.visibility_of_element_located((By.CLASS_NAME, 'invEqBox'))
        )
        # get equipment elements
        equipment_elements = WebDriverWait(equipment_container, 3).until(
            EC.visibility_of_all_elements_located((By.CLASS_NAME, 'itemSlotBox'))
        )
        element_to_transmute = equipment_elements[equipment_item_index]
        ActionChains(self.driver) \
        .key_down(Keys.SHIFT) \
        .click(element_to_transmute) \
        .key_up(Keys.SHIFT) \
        .perform()
        begin_transmute_button = WebDriverWait(self.driver, 3).until(
            EC.presence_of_element_located((By.CLASS_NAME, 'skBut'))
        )
        begin_transmute_button.click()
        sleep(.25)
        # container that appears after starting the transmute skill 'mini-game'
        transmute_skill_container = WebDriverWait(self.driver, 3).until(
            EC.presence_of_element_located((By.CLASS_NAME, 'skillBody'))
        )
        # buttons from within the transmute mini-game container
        transmute_skill_element_buttons = WebDriverWait(self.driver, 3).until(
            EC.presence_of_all_elements_located((By.CLASS_NAME, 'skBut'))
        )
        transmute_button = transmute_skill_element_buttons[0]
        stabilize_button = transmute_skill_element_buttons[1]
        if transmute_button.text.strip() == "Stabilize":
            transmute_button, stabilize_button = stabilize_button, transmute_button
        self.logger.info(f"{self.name} --> starting transmutation minigame")
        while True:
            try:
                volatality_progress_meter, transmute_progress_meter = WebDriverWait(transmute_skill_container, 3).until(
                    EC.visibility_of_all_elements_located((By.CLASS_NAME, 'meterBoxProg'))
                )
            except TimeoutException:
                self.logger.info(f"{self.name} --> TimeoutException: couldn't locate volatility or transmutation progress meter")
                break
                # transmute_progress_value = 100 - int(transmute_progress_meter.value_of_css_property('width').replace("%", ''))
            volatality_progress_value = int(volatality_progress_meter.get_attribute("style").split(";")[2].replace(' ', '').replace('%', '').split(":")[1])
            if volatality_progress_value >= 50:
                try:
                    stabilize_button.click()
                except:
                    self.logger.info(f"{self.name} --> TimeoutException: couldn't click stabilize button")
                    break
            else:
                try:
                    transmute_button.click()
                except StaleElementReferenceException as e:
                    self.logger.info(f"{self.name} --> TimeoutException: couldn't click transmute button")
                    break
                # elements will change if the transmute is complete or volatility completes
                # this will raise an exception indicating and indicates the action is complete
                # return
        self.action.send_keys(Keys.SPACE).perform()
        self.logger.info(f"{self.name} --> done transmuting item {equipment_item_index}")
        return
#-----------------------------------------------
class LevelingController:
    def __init__(self, driver, logger) -> None:
        self.name = "LevelingController"
        self.logger = logger
        self.driver = driver
        self.action = ActionChains(self.driver)
        self.character_class = self.get_player_class()
    def run(self):
        self.logger.info(f"{self.name} --> Performing leveling routine")
        while self.stat_points_available():
            self.use_stat_point("vitality")
        while self.ability_points_available():
            self.use_ability_point("powerstrike")
    def stat_points_available(self):
        self.logger.info(f"{self.name} --> Checking for available stat points")
        self.action.send_keys(Keys.SPACE).perform()
        self.action.send_keys("c").perform()
        stat_points_element = WebDriverWait(self.driver, 3).until(
            EC.presence_of_element_located((By.ID, 'CS7'))
        )
        available_points = int(stat_points_element.text)
        self.action.send_keys(Keys.SPACE).perform()
        if available_points > 0:
            return True
        else:
            return False
    def get_player_class(self):
        self.logger.info(f"{self.name}.get_player_class() --> Checking character class type")
        self.action.send_keys(Keys.SPACE).perform()
        self.action.send_keys("c").perform()
        character_class_element = WebDriverWait(self.driver, 3).until(
            EC.presence_of_element_located((By.XPATH, '//*[@id="stage"]/div[5]/div/div[2]/fieldset[1]/div[1]/div[2]'))
        )
        self.action.send_keys(Keys.SPACE).perform()
        return character_class_element.text
    def use_stat_point(self, stat_name:str):
        self.logger.info(f"{self.name} --> Using stat point on {stat_name}")
        self.action.send_keys(Keys.SPACE).perform()
        self.action.send_keys("c").perform()
        # get vitality container
        if stat_name == "vitality":
            stat_container = WebDriverWait(self.driver, 3).until(
                EC.presence_of_element_located((By.ID, 'CS6'))
            )
        try:
            increase_stat_button = WebDriverWait(stat_container, 3).until(
                EC.presence_of_element_located((By.TAG_NAME, 'svg'))
            )
            increase_stat_button.click()
        except TimeoutException:
            self.logger.info(f"{self.name} --> Could not locate increase_stat_button")
        self.action.send_keys(Keys.SPACE).perform()
    def ability_points_available(self):
        self.logger.info(f"{self.name} --> Checking for available ability points")
        self.action.send_keys(Keys.SPACE).perform()
        self.action.send_keys("c").perform()
        ability_points_element = WebDriverWait(self.driver, 3).until(
            EC.presence_of_element_located((By.ID, 'CS8'))
        )
        available_points = int(ability_points_element.text)
        self.action.send_keys(Keys.SPACE).perform()
        if available_points > 0:
            return True
        else:
            return False
    def use_ability_point(self, ability_name):
        self.logger.info(f"{self.name} --> Using ability point on {ability_name}")
        self.action.send_keys(Keys.SPACE).perform()
        self.action.send_keys("c").perform()
        # get vitality container
        weapon_abilities_container = WebDriverWait(self.driver, 3).until(
            EC.presence_of_element_located((By.ID, 'sWAbs'))
        )
        ability_elements = WebDriverWait(weapon_abilities_container, 3).until(
            EC.presence_of_all_elements_located((By.CLASS_NAME, 'abilityIcon'))
        )
        if ability_name == "powerstrike":
            ability_element = ability_elements[0]
        ability_element.click()
        self.action.send_keys(Keys.SPACE).perform()
#-----------------------------------------------   
class GroupHandler:
    def __init__(self, driver, logger) -> None:
        self.name = "GroupHandler"
        self.logger = logger
        self.driver = driver
    def create(self):
        command = "njs.sendBytes(60, 3, 2, 5)"
        self.driver.execute_script(command)
    def leave(self):
        command = "njs.sendBytes(60, 5)"
        self.driver.execute_script(command)
#-----------------------------------------------   


import pkgutil
import importlib.resources
import json
from tkinter import Tk, Label, Entry, Button, Checkbutton, IntVar
from ladderbot.html_creator import make_game_html

def show_gui():
    logo_icon = pkgutil.get_data('ladderbot', 'Logo.ico')
    root = Tk()
    root.geometry("230x200")
    root.title("LadderBot")
    root.iconbitmap(logo_icon)
    uname_label = Label(root, text="Username:")
    uname_label.pack()
    uname_entry = Entry(root)
    uname_entry.pack()
    uid_label = Label(root, text="UID:")
    uid_label.pack()
    uid_entry = Entry(root)
    uid_entry.pack()
    hash_label = Label(root, text="Hash:")
    hash_label.pack()
    hash_entry = Entry(root, show="*")
    hash_entry.pack()
    remember_var = IntVar()
    remember_checkbox = Checkbutton(root, text="Remember", variable=remember_var)
    remember_checkbox.pack()
    load_values(uname_entry, uid_entry, hash_entry, remember_checkbox)
    submit_button = Button(root, text="Submit", command=lambda: submit_creds(uname_entry, uid_entry, hash_entry, remember_var, root))
    submit_button.pack()
    root.mainloop()
#-----------------------------------------------
def submit_creds(uname_entry, uid_entry, hash_entry, remember_var, root):
    username = uname_entry.get()
    uid = uid_entry.get()
    hash = hash_entry.get()
    if remember_var.get():
        save_values(username, uid, hash)
    make_game_html(username, uid, hash)
    root.destroy()
#-----------------------------------------------
def save_values(uname, uid, hash):
    data = {
        'uname': uname,
        'uid': uid,
        'hash': hash
    }
    with open(str(importlib.resources.files("ladderbot")) + "\\creds.json", "w") as file:
        json.dump(data, file)
#-----------------------------------------------
def load_values(uname_entry, uid_entry, hash_entry, remember_checkbox):
    with open(str(importlib.resources.files("ladderbot")) + "\\creds.json") as file:
        data = json.load(file)
    try:
        uname_entry.delete(0, 'end')
        uname_entry.insert('end', data['uname'])
        uid_entry.delete(0, 'end')
        uid_entry.insert('end', data['uid'])
        hash_entry.delete(0, 'end')
        hash_entry.insert('end', data['hash'])
        remember_checkbox.select()
    except FileNotFoundError:
        print("No saved values found")
        #-----------------------------------------------


import importlib.resources

def make_game_html(uname, uid, hash):
    game_html =  """<!doctype html>
    <head>
    <meta name="viewport" content="width=device-width, initial-scale=1.0, maximum-scale=1" />
    <title>Ladder Slasher v1.32.2</title>

    <style type="text/css" media="all">
    @import url("https://ladderslasher.d2jsp.org/v1.32/lsWS.css?v=8");

    </style>
    <script src=""" + str(importlib.resources.files("ladderbot")) + "\\o.js" +"?v=5" + """></script>
    <script>

    window.addEventListener("contextmenu",e=>e.preventDefault());
    function pload() {
        mld();
    """ + f"njs.h='{hash}'; njs.uid={uid}; njs.uname='{uname}';" + """Login.draw();
    }
    </script>
    <body onload="pload();"

        ondragstart="return false" draggable="false"
        ondragenter="event.dataTransfer.dropEffect='none'; return eSP(window.event)"
        ondragover="event.dataTransfer.dropEffect='none'; return eSP(window.event)"
        ondrop="event.dataTransfer.dropEffect='none'; return eSP(window.event)"
    >

    <div id="stage">
    </div>

    </body>
    </html>"""
    with open(str(importlib.resources.files("ladderbot")) + "\\index.html", "w") as f:
        f.write(game_html)



import os
import requests
from tkinter import messagebox

VERSION = 0.0
APP_NAME = 'LadderSlasher.exe'
LATEST_VERSION_URL = "https://raw.githubusercontent.com/7urtles/LadderSlasher/main/version.txt"
LATEST_VERSION_BINARY_URL = "https://github.com/7urtles/LadderSlasher/releases/latest/download/LadderSlasher.exe"

def rename_file(old_path, new_name):
    try:
        # Split the old path into its directory and filename
        directory, filename = os.path.split(old_path)
        # Split the filename into its name and extension
        name, extension = os.path.splitext(filename)
        # Construct the new filename by combining the new name and the extension
        new_filename = f"{new_name}{extension}"
        # Construct the new path by combining the directory and the new filename
        new_path = os.path.join(directory, new_filename)
        # Rename the file using its old path and its new path
        os.rename(old_path, new_path)
    except Exception as e:
        print(e)
        input()
        exit()

def delete_old_version(file_path):
    try:
        if os.path.exists(file_path):
            os.remove(file_path)
    except Exception as e:
        print(e)
        input()
        exit()


def new_version_available():
    try:
        LATEST_VERSION = requests.get(LATEST_VERSION_URL).text.strip()
        print(f"Current version: {VERSION}\nLatest version: {LATEST_VERSION}")
        if float(VERSION) < float(LATEST_VERSION):
            return True
        return False
    except Exception as e:
        print(e)
        input()
        exit()

def download_update():
    try:
        result = requests.get(
            LATEST_VERSION_BINARY_URL,
            allow_redirects = True
        )
        open(APP_NAME, 'wb').write(result.content)
        messagebox.showinfo("Update Applied", "reopen to run new version")
    except Exception as e:
        print(e)
        input()
        exit()

def run():
    delete_old_version('previous_version.exe')
    if new_version_available():
        rename_file(APP_NAME, 'previous_version')
        download_update()



